package rail

import (
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"log"
	"net/http"
	"net/url"
	"strings"
	"time"

	"github.com/eclipse/paho.golang/autopaho"
	"github.com/eclipse/paho.golang/paho"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promauto"

	"git.majava.org/software/tripla/tripla-fintraffic-rt/pkg/common"
	"git.majava.org/software/tripla/tripla-fintraffic-rt/pkg/util"
)

var (
	activeTargets = promauto.NewGauge(prometheus.GaugeOpts{
		Name: "tripla_fintrafficrt_targets_active",
		Help: "Number of journeys currently being tracked",
	})
	messagesProcessed = promauto.NewCounter(prometheus.CounterOpts{
		Name: "tripla_fintrafficrt_message_processed",
		Help: "The total number of MQTT messages processed",
	})
	messagesInteresting = promauto.NewCounter(prometheus.CounterOpts{
		Name: "tripla_fintrafficrt_message_interesting",
		Help: "The number of interesting MQTT messages that have been forwarded to the backend",
	})
	messageErrors = promauto.NewCounter(prometheus.CounterOpts{
		Name: "tripla_fintrafficrt_message_errors",
		Help: "The total number of errors encountered while processing messages",
	})
	lastMessage = promauto.NewGauge(prometheus.GaugeOpts{
		Name: "tripla_fintrafficrt_message_last_time",
		Help: "Timestamp when the last message was received",
	})
	backendErrors = promauto.NewCounter(prometheus.CounterOpts{
		Name: "tripla_fintrafficrt_backend_errors",
		Help: "The total number of errors encountered while sending data to the backend",
	})
)

type Processor struct {
	common *common.Processor

	trackedJourneys map[string]*TrackedJourney
}

func NewProcessor(baseUrl string, apiToken string) *Processor {
	return &Processor{
		common:          common.NewProcessor(baseUrl, apiToken),
		trackedJourneys: map[string]*TrackedJourney{},
	}
}

func (p *Processor) updateJourneys() {
	journeys := []TrackedJourney{}
	err := p.common.GetAndParseJson(
		fmt.Sprintf("/internal/live/journeys-to-track/%v", DataSourceType),
		&journeys,
	)
	if err != nil {
		backendErrors.Inc()
		panic(err)
	}

	journeyMap := map[string]*TrackedJourney{}
	for _, journey := range journeys {
		journey.VersionNumber = 0
		journeyMap[journey.JourneyLocator.ToString()] = &journey
	}

	log.Printf("Now tracking %v journeys", len(journeyMap))
	p.trackedJourneys = journeyMap
	activeTargets.Set(float64(len(journeyMap)))
}

func (p *Processor) processCompositionMessage(msg *paho.Publish) error {
	var m MqttCompositionMessage
	if err := json.Unmarshal(msg.Payload, &m); err != nil {
		log.Printf("Message could not be parsed (%s): %s\n", msg.Payload, err)
		return nil
	}

	trainId := m.TrainId()
	target, ok := p.trackedJourneys[trainId.ToString()]
	if !ok {
		log.Printf("Received message for not-interesting train %v\n", trainId.ToString())
		return nil
	}

	if target.VersionNumber >= m.Version {
		log.Printf(
			"Received message with duplicate version ID: %v %v %v >= %v\n",
			target.JourneyId, trainId.ToString(), target.VersionNumber, m.Version)
		return nil
	}

	messagesInteresting.Inc()
	log.Printf("Received message (%v): %v\n", target.JourneyId, m)

	go func() {
		data, err := json.Marshal(map[string]any{
			"journeyId": target.JourneyId,
			"sections":  m.JourneySections,
		})
		if err != nil {
			log.Printf("Failed to marshal HTTP POST body: %v", err)
			return
		}

		req, err := http.NewRequest(
			"POST",
			fmt.Sprintf("%v/internal/live/process/fintraffic-composition", p.common.BaseUrl),
			bytes.NewBuffer(data),
		)
		if err != nil {
			log.Printf("Failed to create request: %v", err)
			return
		}

		req.Header.Set("Authorization", fmt.Sprintf("Bearer %v", p.common.ApiToken))
		req.Header.Set("User-Agent", p.common.UserAgent())
		req.Header.Set("Content-Type", "application/json")

		response, err := p.common.HttpClient.Do(req)
		if err != nil {
			backendErrors.Inc()
			log.Printf("Failed to execute request: %v", err)
		}
		defer response.Body.Close()

		if response.StatusCode != http.StatusNoContent {
			backendErrors.Inc()
			log.Printf("Unexpected status %v %v", response.StatusCode, response.Status)
		}
	}()

	target.VersionNumber = m.Version

	return nil
}

func (p *Processor) processMqttMessage(msg *paho.Publish) error {
	if strings.HasPrefix(msg.Topic, "compositions/") {
		return p.processCompositionMessage(msg)
	}

	return errors.New(fmt.Sprintf("Unknown topic '%v'", msg.Topic))
}

func (p *Processor) processMqttMessages() {
	u, err := url.Parse("mqtt://rata-mqtt.digitraffic.fi:1883")
	if err != nil {
		panic(err)
	}

	ctx := context.Background()

	cliCfg := autopaho.ClientConfig{
		ServerUrls:                    []*url.URL{u},
		KeepAlive:                     20,
		CleanStartOnInitialConnection: true,
		SessionExpiryInterval:         60,
		OnConnectionUp: func(cm *autopaho.ConnectionManager, connAck *paho.Connack) {
			if _, err := cm.Subscribe(ctx, &paho.Subscribe{
				Subscriptions: []paho.SubscribeOptions{
					{Topic: "compositions/#", QoS: 1},
				},
			}); err != nil {
				log.Printf("Failed to subscribe (%s). Probably due to connection drop so will retry\n", err)
				return
			}

			log.Println("MQTT connection established")
		},
		OnConnectError: func(err error) { log.Printf("Error whilst attempting connection: %s\n", err) },
		Errors:         log.Default(),

		ClientConfig: paho.ClientConfig{
			ClientID: fmt.Sprintf("tripla-rt-%v", util.RandStringRunes(8)),
			OnPublishReceived: []func(paho.PublishReceived) (bool, error){
				func(pr paho.PublishReceived) (bool, error) {
					messagesProcessed.Inc()
					lastMessage.SetToCurrentTime()

					err = p.processMqttMessage(pr.Packet)
					if err != nil {
						log.Printf("Failed to process a message: %v\n", err)
						messageErrors.Inc()
					}
					return true, nil
				},
			},
			OnClientError: func(err error) { log.Printf("MQTT client error: %s\n", err) },
			OnServerDisconnect: func(d *paho.Disconnect) {
				if d.Properties != nil {
					log.Printf("Server requested disconnect: %s\n", d.Properties.ReasonString)
				} else {
					log.Printf("Server requested disconnect; reason code: %d\n", d.ReasonCode)
				}
			},
		},
	}

	_, err = autopaho.NewConnection(ctx, cliCfg)
	if err != nil {
		panic(err)
	}
}

func (p *Processor) StartJourneyUpdater() {
	p.updateJourneys()
	go func() {
		for {
			time.Sleep(time.Minute)
			p.updateJourneys()
		}
	}()

	p.processMqttMessages()
}
