package proxy

import (
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"log/slog"
	"net/http"
	"net/url"
	"os"
	"sync"
	"time"

	"git.majava.org/software/tripla/tripla-gtfs-proxy/pkg/config"
)

type Proxy struct {
	triplaBaseUrl  string
	triplaApiToken string
	httpClient     http.Client

	cfg *config.Config

	sourceConfig *map[string]*GtfsSourceConfig
	sources      map[string]*GtfsSource
}

func (p *Proxy) UserAgent() string {
	return fmt.Sprintf("tripla-fintraffic-rt/%v", os.Getenv("TRIPLA_VERSION"))
}

func (p *Proxy) GetSource(source string) (*GtfsSource, error) {
	s, ok := p.sources[source]
	if ok {
		return s, nil
	}

	cfg, err := p.getSourceConfig()
	if err != nil {
		return nil, err
	}

	sConfig, ok := (*cfg)[source]
	if !ok {
		return nil, errors.New("unknown source")
	}

	src := GtfsSource{
		name:         source,
		config:       sConfig,
		proxy:        p,
		downloadLock: &sync.Mutex{},
		httpClient: http.Client{
			Timeout: 20 * time.Second,
		},
	}

	if sConfig.overrides.Proxy != "" {
		proxy, err := url.Parse(sConfig.overrides.Proxy)
		if err != nil {
			return nil, err
		}

		src.httpClient.Transport = &http.Transport{
			Proxy: http.ProxyURL(proxy),
		}
	}

	p.sources[source] = &src
	return p.sources[source], nil
}

func (p *Proxy) getSourceConfig() (*map[string]*GtfsSourceConfig, error) {
	if p.sourceConfig != nil {
		return p.sourceConfig, nil
	}

	sourceConfig := map[string]*GtfsSourceConfig{}

	configs := []*GtfsSourceConfig{}
	err := p.GetAndParseJson("/internal/gtfs/sources", &configs)
	if err != nil {
		return nil, err
	}

	for _, cfg := range configs {
		cfg.overrides = &config.SourceConfig{}
		sourceConfig[cfg.Id] = cfg
	}

	for _, overrides := range p.cfg.Sources {
		if scfg, ok := sourceConfig[overrides.Name]; ok {
			scfg.overrides = overrides
		} else {
			slog.Warn("Overrides present for unknown source",
				"source", overrides.Name)
		}
	}

	p.sourceConfig = &sourceConfig
	return p.sourceConfig, nil
}

func (p *Proxy) GetAndParseJson(path string, target any) error {
	req, err := http.NewRequest("GET", fmt.Sprintf("%v%v", p.triplaBaseUrl, path), nil)
	if err != nil {
		return err
	}
	req.Header.Set("Authorization", fmt.Sprintf("Bearer %v", p.triplaApiToken))
	req.Header.Set("User-Agent", p.UserAgent())

	response, err := p.httpClient.Do(req)
	if err != nil {
		return err
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		return errors.New(fmt.Sprintf("Unexpected status %v %v", response.StatusCode, response.Status))
	}

	body, err := io.ReadAll(response.Body)
	if err != nil {
		return err
	}

	return json.Unmarshal(body, target)
}

func NewProxy(triplaBaseUrl string, triplaApiToken string, cfg *config.Config) *Proxy {
	return &Proxy{
		cfg:            cfg,
		triplaBaseUrl:  triplaBaseUrl,
		triplaApiToken: triplaApiToken,
		httpClient:     http.Client{},
		sources:        map[string]*GtfsSource{},
	}
}
