package proxy

import (
	"archive/zip"
	"bytes"
	"errors"
	"io"
	"log/slog"
	"net/http"
	"sync"
	"time"
)

type GtfsSource struct {
	name   string
	config *GtfsSourceConfig
	proxy  *Proxy

	httpClient   http.Client
	latestData   *cachedData
	downloadLock *sync.Mutex
}

type GtfsData map[string]CsvFile

type cachedData struct {
	data    GtfsData
	etag    string
	expires time.Time
}

func (s *GtfsSource) fetchContent() (*cachedData, error) {
	req, err := http.NewRequest("GET", s.config.Url, nil)
	if err != nil {
		return nil, err
	}
	req.Header.Set("User-Agent", s.proxy.UserAgent())

	if s.latestData != nil && s.latestData.etag != "" {
		req.Header.Set("If-None-Match", s.latestData.etag)
	}

	response, err := s.httpClient.Do(req)
	if err != nil {
		return nil, err
	}
	defer response.Body.Close()

	if response.StatusCode == http.StatusNotModified {
		slog.Info("Data has not changed, using cached data",
			"source", s.name,
			"etag", s.latestData.etag,
		)
		return s.latestData, nil
	}

	b, err := io.ReadAll(response.Body)
	if err != nil {
		return nil, err
	}

	zf, err := zip.NewReader(bytes.NewReader(b), int64(len(b)))
	if err != nil {
		return nil, err
	}

	files := GtfsData{}
	for _, file := range zf.File {
		rows, err := parseZipFile(file)
		if err != nil {
			return nil, err
		}

		files[file.Name] = *rows
	}

	cached := cachedData{
		data:    files,
		etag:    response.Header.Get("ETag"),
		expires: time.Now().Add(time.Hour),
	}

	// TODO: Parse expires header and similar

	slog.Info("Downloaded data",
		"source", s.name,
		"etag", cached.etag,
		"expires", cached.expires,
	)

	return &cached, nil
}

func (s *GtfsSource) getContent() (*cachedData, error) {
	if s.latestData != nil {
		if s.latestData.expires.Before(time.Now()) {
			slog.Info("Data has expired, fetching updates",
				"source", s.name,
				"etag", s.latestData.etag,
			)
		} else {
			return s.latestData, nil
		}
	}

	s.downloadLock.Lock()
	defer s.downloadLock.Unlock()
	if s.latestData != nil {
		// Data was downloaded while we were waiting for the lock,
		// use newly discovered data instead of re-downloading it.
		return s.latestData, nil
	}

	slog.Info("Downloading data",
		"source", s.name,
	)

	data, err := s.fetchContent()
	if err != nil {
		if s.latestData != nil {
			slog.Error("Failed to download new data, using cached data",
				"source", s.name,
				"error", err,
			)
			// TODO: respect retry-after
			s.latestData.expires = time.Now().Add(time.Minute * 15)
			return s.latestData, nil
		}

		return nil, err
	}

	s.latestData = data
	return s.latestData, nil
}

func (s *GtfsSource) GetFile(file string) (*CsvFile, error) {
	content, err := s.getContent()
	if err != nil {
		return nil, err
	}

	data, ok := (*content).data[file]
	if !ok {
		return nil, errors.New("no such file")
	}

	return &data, nil
}
