package server

import (
	"encoding/json"
	"fmt"
	"io"
	"log"
	"net/http"

	"github.com/prometheus/client_golang/prometheus/promhttp"

	"git.majava.org/software/tripla/tripla-gtfs-proxy/pkg/proxy"
)

type ProxyServer struct {
	Proxy *proxy.Proxy
}

func (ps *ProxyServer) getFile(w http.ResponseWriter, r *http.Request) {
	source, err := ps.Proxy.GetSource(r.PathValue("source"))
	if err != nil {
		w.WriteHeader(500)
		w.Write([]byte("Failed to get source!\n"))
		log.Printf("Failed to get source: %v", err)
		return
	}

	body := struct {
		Filter    *map[string][]string `json:"filter"`
		UniqueKey *string              `json:"uniqueKey"`
	}{}
	if r.Method == "QUERY" && r.ContentLength >= 1 {
		data, err := io.ReadAll(r.Body)
		if err != nil {
			w.WriteHeader(400)
			w.Write([]byte("Failed to read body!\n"))
			log.Printf("Failed to read body: %v", err)
			return
		}

		err = json.Unmarshal(data, &body)
		if err != nil {
			w.WriteHeader(400)
			w.Write([]byte("Failed to parse body!\n"))
			log.Printf("Failed to parse body: %v", err)
			return
		}
	}

	content, err := source.GetFile(fmt.Sprintf("%v.txt", r.PathValue("file")))
	if err != nil {
		w.WriteHeader(500)
		w.Write([]byte("Failed to get content!\n"))
		log.Printf("Failed to get content: %v", err)
		return
	}

	if body.Filter != nil {
		filterMap := map[string]map[string]bool{}
		for key, value := range *body.Filter {
			filters := map[string]bool{}
			for _, s := range value {
				filters[s] = true
			}
			filterMap[key] = filters
		}

		content.Values = filter(
			content.Values,
			func(entry []string) bool {
				for key, expected := range filterMap {
					value := entry[content.ColumnIndexes[key]]
					_, ok := expected[value]
					if !ok {
						return false
					}
				}

				return true
			},
		)
	}

	var result any
	if body.UniqueKey != nil {
		values := map[string]bool{}
		for _, row := range content.Values {
			value := row[content.ColumnIndexes[*body.UniqueKey]]
			values[value] = true
		}

		keys := make([]string, len(values))

		i := 0
		for k := range values {
			keys[i] = k
			i++
		}

		result = keys
	} else {
		result = content.ToJson()
	}

	bytes, err := json.Marshal(result)
	if err != nil {
		w.WriteHeader(500)
		w.Write([]byte("Failed to marshal content!\n"))
		log.Printf("Failed to marshal content: %v", err)
		return
	}

	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(200)
	w.Write(bytes)
}

func (ps *ProxyServer) GetHandler() http.Handler {
	mux := http.NewServeMux()
	mux.Handle("/metrics", promhttp.Handler())
	mux.HandleFunc("/gtfs/v1/{source}/{file}", ps.getFile)

	return mux
}
