package processor

import (
	"fmt"
	"log"
	"net/http"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promauto"
)

var (
	activeTargets = promauto.NewGauge(prometheus.GaugeOpts{
		Name: "tripla_gtfsrt_targets_active",
		Help: "Number of journeys currently being tracked",
	})
	backendErrors = promauto.NewCounter(prometheus.CounterOpts{
		Name: "tripla_gtfsrt_backend_errors",
		Help: "The total number of errors encountered while sending data to the backend",
	})
)

type Processor struct {
	baseUrl    string
	apiToken   string
	httpClient http.Client
	sources    []*GtfsRtSource
}

func NewProcessor(baseUrl string, apiToken string) *Processor {
	return &Processor{
		baseUrl:    baseUrl,
		apiToken:   apiToken,
		httpClient: http.Client{},
	}
}

func (p *Processor) StartSources(sources []*GtfsRtSource) {
	p.sources = sources

	for _, source := range sources {
		go func() {
			p.StartSource(source)
		}()
	}
}

func (p *Processor) updateJourneys() error {
	journeys := []TrackedJourney{}
	err := p.GetAndParseJson(fmt.Sprintf("/internal/live/journeys-to-track/%v", DataSourceType), &journeys)
	if err != nil {
		backendErrors.Inc()
		return err
	}

	activeTargets.Set(float64(len(journeys)))

	journeysBySource := map[string]map[string]TrackedJourney{}
	for _, journey := range journeys {
		source := journey.SourceId.ToString()

		elem, ok := journeysBySource[source]
		if !ok {
			elem = map[string]TrackedJourney{}
			journeysBySource[source] = elem
		}

		elem[journey.JourneyLocator.TripId] = journey
	}

	for _, source := range p.sources {
		if sourceJourneys, ok := journeysBySource[source.Identifier.ToString()]; ok {
			log.Printf("Found %v journeys for %v", len(sourceJourneys), source.Identifier.ToString())
			source.journeysToTrack = sourceJourneys
		} else {
			log.Printf("Found no journeys for %v", source.Identifier.ToString())
			source.journeysToTrack = map[string]TrackedJourney{}
		}
	}

	return nil
}

func (p *Processor) StartJourneyUpdater() {
	go func() {
		for {
			err := p.updateJourneys()

			if err != nil {
				log.Printf("Failed to update journeys: %v", err)
			}

			time.Sleep(time.Minute)
		}
	}()
}
