package processor

import (
	"errors"
	"fmt"
	"io"
	"log"
	"net/http"
	"time"

	"github.com/pquerna/cachecontrol"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promauto"
	"google.golang.org/protobuf/proto"

	"git.majava.org/software/tripla/tripla-gtfs-rt/internal/gtfs/rt"
)

var (
	sourceErrors = promauto.NewCounterVec(prometheus.CounterOpts{
		Name: "tripla_gtfsrt_source_errors",
		Help: "The total number of errors encountered while processing data",
	}, []string{"source"})
	sourceLastTimestamp = promauto.NewGaugeVec(prometheus.GaugeOpts{
		Name: "tripla_gtfsrt_source_latest_version_time",
		Help: "Timestamp of last GTFS-RT version as indicated in the data",
	}, []string{"source"})
)

const DefaultDelay = time.Second * 10

func (p *Processor) processJourneys(s *GtfsRtSource) (time.Duration, error) {
	if len(s.journeysToTrack) == 0 {
		log.Printf("%v: no journeys to track\n", s.Identifier.ToString())
		return DefaultDelay, nil
	}

	req, err := http.NewRequest("GET", s.GtfsRtUrl, nil)
	if err != nil {
		return DefaultDelay, err
	}

	response, err := p.httpClient.Do(req)
	if err != nil {
		sourceErrors.With(map[string]string{"source": s.Identifier.ToString()}).Inc()
		return DefaultDelay, err
	}
	defer response.Body.Close()

	_, expires, _ := cachecontrol.CachableResponse(req, response, cachecontrol.Options{})
	delay := DefaultDelay
	if !expires.IsZero() {
		delay = expires.Sub(time.Now()) + time.Second
	}

	if response.StatusCode != http.StatusOK {
		sourceErrors.With(map[string]string{"source": s.Identifier.ToString()}).Inc()
		return delay, errors.New(fmt.Sprintf("Unexpected status %v %v", response.StatusCode, response.Status))
	}

	body, err := io.ReadAll(response.Body)
	if err != nil {
		sourceErrors.With(map[string]string{"source": s.Identifier.ToString()}).Inc()
		return DefaultDelay, err
	}

	message := rt.FeedMessage{}
	err = proto.Unmarshal(body, &message)
	if err != nil {
		sourceErrors.With(map[string]string{"source": s.Identifier.ToString()}).Inc()
		return delay, err
	}

	log.Printf("%v: got RT data with timestamp %v", s.Identifier.ToString(), message.GetHeader().GetTimestamp())
	sourceLastTimestamp.
		With(map[string]string{"source": s.Identifier.ToString()}).
		Set(float64(message.GetHeader().GetTimestamp()))

	for _, entity := range message.GetEntity() {
		tripId := entity.GetTripUpdate().GetTrip().GetTripId()
		journey, ok := s.journeysToTrack[tripId]
		if !ok {
			continue
		}

		log.Printf("%v: %v => %v\n", s.Identifier.ToString(), tripId, journey.JourneyId)
		for _, update := range entity.GetTripUpdate().GetStopTimeUpdate() {
			arrival := int32(0)
			if update.GetArrival() != nil {
				arrival = update.GetArrival().GetDelay()
			}

			departure := int32(0)
			if update.GetDeparture() != nil {
				departure = update.GetDeparture().GetDelay()
			}

			log.Printf(" %v %v %v", update.GetStopId(), arrival, departure)
		}
	}

	return delay, err
}

func (p *Processor) StartSource(s *GtfsRtSource) {
	// Init metrics as 0
	sourceErrors.With(map[string]string{"source": s.Identifier.ToString()})
	sourceLastTimestamp.With(map[string]string{"source": s.Identifier.ToString()})

	for {
		delay, err := p.processJourneys(s)

		if err != nil {
			log.Printf("Failed to process %v: %v", s.Identifier.ToString(), err)
		}

		time.Sleep(delay)
	}
}
